# -*- coding: utf-8 -*-

"""
    File name: reservoir_net.py
    Description: a neural network with readout units and a feedback loop
    Authors: Roman Pogodin, adapted from Moritz Deger and Joseph Heng
    Comments: taken from https://github.com/romanpogodin/reservoir-attractor-learning
    Python version: 3.6
"""

import numpy as np
from warnings import warn


def correlate_multidim(x, y):
    """
    Correlation of two periodic signals of the same length, computed with the circular shift.
    x could be multi-dimensional, in which case correlation is performed along the last axis.
    """
    y = np.hstack((y, y))[:-1]
    old_shape = x.shape
    x = x.reshape(-1, x.shape[-1])
    new_x = np.zeros(x.shape)

    for i in range(new_x.shape[0]):
        new_x[i] = np.correlate(x[i], y, mode='valid')

    return new_x.reshape(old_shape)


def cross_correlate_periodic(x, y):
    """
    Normalised cross correlation of two periodic signals of the same length, computed with the circular shift.
    x could be multi-dimensional, in which case correlation is performed along the last axis.
    """
    correlation = correlate_multidim(x - np.expand_dims(np.mean(x, axis=-1), axis=-1), y - np.mean(y))
    std = np.expand_dims(np.std(x, axis=-1), axis=-1) * np.std(y) * len(y)
    return correlation / (std + 1e-16 * (std == 0.0))  # will return zero along the axis with std=0


def compute_shifted_cross_correlation(data, targets):
    """
    Compute cross-correlation with a temporal shift.

    Shift the signal to achieve better correlation (as it might be naturally shifted due to imperfectly learnt signal).
    Matches the MSE from
    Hoerzer, G. M.; Legenstein, R. & Maass, W.
    Emergence of Complex Computational Structures From Chaotic Neural Networks
    Through Reward-Modulated Hebbian Learning. Cereb Cortex, 2012
    :param data:        input signal of the shape [..., len(targets)]
    :param targets:     target signal
    :return: Minimum normalised cross-correlation of the shape data.shape[:-1]
    """
    cross_corr = cross_correlate_periodic(data, targets)

    return np.max(cross_corr, axis=-1)


class LearningRule:
    def __init__(self):
        self._time = 0

    def compute_update(self, f, z, r):
        return np.zeros((len(z), len(r)))


class ReservoirNet(object):
    """
    A recurrent neural net with trainable output units. Variable names are chosen to be consistent
    with the reference paper.
    This is an implementation of
    Hoerzer, G. M.; Legenstein, R. & Maass, W.
    Emergence of Complex Computational Structures From Chaotic Neural Networks
    Through Reward-Modulated Hebbian Learning. Cereb Cortex, 2012
    """

    def __init__(self, learning_rule=LearningRule(), num_rec=1000, num_out=1, randseed=None, chaos_constant=1.5):
        """

        :param learning_rule:           a class w/ compute_update(f, z, r), return weights w/ shape(np.outer(z, r))
        :param num_rec:                 number of recurrent neurons
        :param num_out:                 number of output units (== target function dimension)
        :param randseed:                random seed for all random operations
        :param chaos_constant:          strength of connectivity, regulates chaos in the net
        """
        self._num_rec = num_rec
        self._num_out = num_out
        self._theta_state = 0.01  # amplitude of firing rates noise
        self._rec_connect_prob = 0.1  # probability of a recurrent connection
        self._tau = 50.0  # membrane time constant, ms
        self._dt = 1.0  # time step, ms
        self._lambda = chaos_constant  # strength of recurrent connection, controls the chaotic regime

        self._rand_generator = np.random.RandomState(randseed)
        self._amplitude_fb_weights = 1.0
        self._external_noise_amplitude = 0.05  # Gaussian noise amplitude for input signals

        self.x = np.zeros(self._num_rec)  # membrane potential
        self.r = np.zeros(self._num_rec)  # firing rate
        self.z = np.zeros(self._num_out)  # noisy output
        self.weights_out = self._make_weights_out()  # output weights (plastic)

        self._xi_state = np.zeros(self._num_rec)  # firing rates noise
        self._weights_rec = self._make_weights_rec()  # recurrent weights
        self._weights_fb = self._make_weights_fb()  # feedback weights
        self._dynamics_decay = self._dt / self._tau
        self._external_input = np.zeros(self._num_rec)

        self._learning_rule = learning_rule

        self._memorised_state = None  # used to memorise the state after training

    def _make_indicator_rec(self):
        """
        Creates a binary recurrent connectivity matrix.
        :return: a (_num_rec, _num_rec) 0/1 matrix
        """
        size = (self._num_rec, self._num_rec)
        return self._rand_generator.binomial(1, self._rec_connect_prob, size=size)

    def _make_weights_rec(self):
        """
        Makes random normal recurrent weights.
        :return: a (_num_rec, _num_rec) real matrix
        """
        size = (self._num_rec, self._num_rec)
        std = np.sqrt(1.0 / (self._rec_connect_prob * self._num_rec))
        return self._rand_generator.normal(loc=0.0, scale=std, size=size) * self._make_indicator_rec()

    def _make_weights_fb(self):
        """
        Makes random uniform feedback weights.
        :return: a (_num_rec, _num_out) real matrix
        """
        return self._rand_generator.uniform(low=-self._amplitude_fb_weights,
                                            high=self._amplitude_fb_weights,
                                            size=(self._num_rec, self._num_out))

    def _make_weights_out(self):
        """
        Sets initial output weights to zero.
        :return: a zero (_num_out, _num_rec) matrix
        """
        return np.zeros((self._num_out, self._num_rec))

    def _get_dx(self):
        """Return update of the membrane potential."""
        return self._dynamics_decay * (-self.x + self._lambda * np.dot(self._weights_rec, self.r) +
                                       np.dot(self._weights_fb, self.z) + self._external_input)

    def _set_xi_state(self):
        """Sets noise in firing rates"""
        self._xi_state = self._rand_generator.uniform(low=-self._theta_state,
                                                      high=self._theta_state, size=self._num_rec)

    def _set_r(self):
        """Sets firing rates (including noise)"""
        self.r = np.tanh(self.x) + self._xi_state

    def _set_z(self):
        """Sets activity in readout neurons (including noise)"""
        self.z = np.dot(self.weights_out, self.r)

    def _update_external_input(self, external_input=None):
        """Updates external input and adds noise (if None, sets it to zero)."""
        if external_input is not None:  # add noise
            self._external_input = external_input + self._external_noise_amplitude * \
                                   self._rand_generator.normal(loc=0.0, scale=1.0, size=(self._num_rec,))
        else:
            self._external_input.fill(0)

    def _update_activity(self):
        """Updates activity of the network."""
        self.x += self._get_dx()
        self._set_xi_state()
        self._set_r()
        self._set_z()

    def _update_plasticity(self, target_function):
        """
        Updates plasticity of the network.
        Must be called after _update_activity().
        :param target_function: a (_num_out) array (or a scalar for _num_out=1)
        :return: None
        """
        self.weights_out += self._learning_rule.compute_update(np.array(target_function, ndmin=1),
                                                               self.z, self.r)

    def reset_network_activity(self, x_init=None):
        """
        Sets dynamics of the network to zero.
        :return: None
        """
        if x_init is not None:
            self.x = x_init.copy()
        else:
            self.x = np.zeros(self._num_rec)
        self.r = np.zeros(self._num_rec)
        self.z = np.zeros(self._num_out)

    def memorise_state(self):
        """Saves the network's state. Does not save plasticity parameters."""
        self._memorised_state = (self.x.copy(), self.r.copy(), self.z.copy())

    def recover_memorised_state(self):
        """Recovers the saved network's state."""
        if self._memorised_state is not None:
            self.x, self.r, self.z = self._memorised_state
            self._memorised_state = None
        else:
            warn("Nothing to recover")

    def update(self, external_input=None, target_function=None):
        """
        Updates the state of the network and the output weights (if target_function is not None).
        :param external_input:  None or a (_num_rec) vector. In the latter case Gaussian noise will be added.
        :param target_function: None (for no plasticity) or a (_num_out) array.
        :return: None
        """
        self._update_external_input(external_input)
        self._update_activity()

        if target_function is not None:
            self._update_plasticity(target_function)


class ReservoirInterface:
    """
    An interface for training and testing of a reservoir network.
    """

    def __init__(self, learning_rule, period, n_reservoir_neurons=1000, randseed=42, chaos_constant=1.5, num_out=1):
        self.reservoir = ReservoirNet(
            num_rec=n_reservoir_neurons, randseed=randseed, learning_rule=learning_rule,
            num_out=num_out, chaos_constant=chaos_constant)

        self._period = int(period)
        self._recorded_variables = []
        self._n_target_functions = num_out
        self._warm_up_time = 0  # for reset_network

    def reset_network(self, x_init=None):
        """Resets network and makes it run with no input for self._warm_up_time"""
        self.reservoir.reset_network_activity(x_init)

        for net_time in range(self._warm_up_time):
            self.reservoir.update()

    def _simulate_period(self, external_input=None, target_function=None):
        """
        Runs the network for one period.
        :param external_input:      an already converted (of shape (n_rec,) attractor input or None
        :param target_function:     a target function for train or None for test
        :return: values of each variable from self._recorded_variables of the shape [n_vars, time]
        """
        result = np.zeros((len(self._recorded_variables), self._period))

        current_external_input = None
        current_target_function = None

        for net_time in range(self._period):
            if external_input is not None:
                current_external_input = external_input[net_time]
            if target_function is not None:
                current_target_function = target_function[net_time]

            self.reservoir.update(external_input=current_external_input, target_function=current_target_function)

            for i in range(len(self._recorded_variables)):
                result[i, net_time] = eval('self.reservoir.' + self._recorded_variables[i])

        return result

    def _simulate(self, n_epochs, recorded_variables=None, period_reset=False,
                  external_input=None, target_functions=None):
        """
        Simulates a network for several epochs.
        :param n_epochs:            number of periods for each target function
        :param recorded_variables:  a list of 1d variables to record
        :param period_reset:        whether to reset the net after each period
        :param external_input:      an already converted (of shape (period, n_rec) attractor input or None
        :param target_functions:    a list of target functions for train or None for test
        :return: values of each variable from recorded_variables of the shape [n_vars, time]
        """
        if external_input is not None and len(external_input.shape) == 2:  # for one task
            external_input = external_input[None, :, :]

        if recorded_variables is None:
            self._recorded_variables = []
        else:
            self._recorded_variables = recorded_variables.copy()

        result = np.zeros((len(self._recorded_variables), self._period * n_epochs * self._n_target_functions))

        current_external_input = None
        current_target_function = None

        for i in range(n_epochs * self._n_target_functions):
            if period_reset:
                self.reset_network()

            if external_input is not None:
                current_external_input = external_input[i % self._n_target_functions]
            if target_functions is not None:
                current_target_function = target_functions[i % self._n_target_functions]

            result[:, i * self._period:(i + 1) * self._period] = \
                self._simulate_period(external_input=current_external_input,
                                      target_function=current_target_function)

        return result

    def train(self, n_epochs, target_functions, recorded_variables=None,
              period_reset=False, external_input=None):
        """
        Trains the network for several epochs.
        :param n_epochs:            number of periods for each target function
        :param target_functions:    a list of target functions for train or None for test
        :param recorded_variables:  a list of 1d variables to record
        :param period_reset:        whether to reset the net after each period
        :param external_input:      an already converted (of shape (period, n_rec) attractor input or None
        :return: values of each variable from recorded_variables of the shape [n_vars, time]
        """
        if len(target_functions.shape) == 1:  # for one task
            target_functions = target_functions[None, :]
        self._n_target_functions = target_functions.shape[0]

        return self._simulate(n_epochs, recorded_variables, period_reset, external_input,
                              target_functions)

    def test(self, n_epochs, recorded_variables=None, period_reset=False, external_input=None):
        """
        Test the network for several epochs (no plasticity).
        :param n_epochs:                number of periods for each target function
        :param recorded_variables:      a list of 1d variables to record
        :param period_reset:            whether to reset the net after each period
        :param external_input:          an already converted (of shape (n_rec,) attractor input or None
        :return: values of each variable from recorded_variables of the shape [n_vars, time]
        """
        self.reservoir.memorise_state()
        result = self._simulate(n_epochs, recorded_variables, period_reset, external_input)
        self.reservoir.recover_memorised_state()
        return result

    def run_reservoir(self, target_values, n_epochs, test_frequency, to_record,
                      x_init=None, period_reset=False, external_input=None):
        """
        Runs the reservoir for several training periods, testing it with a fixed frequency.
        :param target_values:
        :param n_epochs:        number of training periods
        :param test_frequency:  how often to perform one-period tests
        :param to_record:       list of variables to record
        :param x_init:          initial value for the membrane potentials
        :param period_reset:    whether to reset activity after each period
        :param external_input:  external input
        :return: dictionaries for recorded variables on train, test, recorded cross-correlation on test
        """
        if not any(x in 'z' for x in to_record):
            to_record.append('z')
        
        period = len(target_values)
        n_tests = 1 + int(np.floor((n_epochs // test_frequency)))
        train_recordings = np.zeros((len(to_record), n_epochs, period))
        test_recordings = np.zeros((len(to_record), n_tests, period))
        correlation = np.zeros(n_tests)  # initial correlation is 0 as w(0)=0

        self.reset_network(x_init)
        test_recordings[:, 0] = self.test(1, to_record, period_reset, external_input)

        for epoch in range(n_epochs):
            train_recordings[:, epoch, :] = self.train(1, target_values, to_record, period_reset, external_input)

            if epoch % test_frequency == 0:
                test_recordings[:, 1 + epoch // test_frequency] = self.test(1, to_record, period_reset, external_input)
                correlation[1 + epoch // test_frequency] = compute_shifted_cross_correlation(
                    test_recordings[0, 1 + epoch // test_frequency, :], target_values)

        train_dict = {}
        test_dict = {}

        for i, variable in enumerate(to_record):
            train_dict[variable] = train_recordings[i]
            test_dict[variable] = test_recordings[i]

        return train_dict, test_dict, correlation
